use crate::gen_random;
use rocket::Request;
use rocket::Response;
use rocket::http::Header;
use rocket::http::Status;
use rocket::response::Responder;
use std::io::Cursor;
use std::io::Read;
use std::io::Seek;
use std::os::unix::fs::FileExt;
use std::os::unix::fs::MetadataExt;

// TODO: Implement file based response

pub struct Data {
    file: Option<String>,
    raw: Option<Vec<u8>>,
}

impl Data {
    pub fn raw(data: Vec<u8>) -> Self {
        Self {
            file: None,
            raw: Some(data),
        }
    }

    pub fn file(path: &str) -> Self {
        Self {
            file: Some(path.to_string()),
            raw: None,
        }
    }

    pub fn len(&self) -> usize {
        if let Some(raw) = &self.raw {
            raw.len()
        } else {
            std::fs::File::open(self.file.as_ref().unwrap())
                .unwrap()
                .metadata()
                .unwrap()
                .size() as usize
        }
    }

    pub fn index(&self, start: usize, end: usize) -> Vec<u8> {
        if let Some(raw) = &self.raw {
            raw[start..=end].to_vec()
        } else {
            let mut file = std::fs::File::open(self.file.as_ref().unwrap()).unwrap();
            file.seek(std::io::SeekFrom::Start(start as u64)).unwrap();
            let mut buf = vec![0; end - start + 1];
            file.read_exact(&mut buf).unwrap();
            buf
        }
    }

    pub fn full(self) -> Result<Vec<u8>, tokio::fs::File> {
        if let Some(raw) = self.raw {
            Ok(raw)
        } else {
            let file = std::fs::File::open(self.file.unwrap()).unwrap();
            let file = tokio::fs::File::from_std(file);
            return Err(file);
        }
    }
}

pub struct DataResponse {
    data: Data,
    content_type: String,
    // TODO : Implement better cache control
    cache_duration: Option<u64>,
}

impl DataResponse {
    #[must_use]
    pub fn new(data: Vec<u8>, content_type: String, cache_duration: Option<u64>) -> Self {
        Self {
            data: Data::raw(data),
            content_type,
            cache_duration,
        }
    }

    #[must_use]
    pub fn new_file(path: &str, content_type: String, cache_duration: Option<u64>) -> Self {
        Self {
            data: Data::file(path),
            content_type,
            cache_duration,
        }
    }
}

#[rocket::async_trait]
impl<'r> Responder<'r, 'static> for DataResponse {
    fn respond_to(self, req: &'r Request<'_>) -> rocket::response::Result<'static> {
        // Handle Range requests
        if let Some(range) = req.headers().get_one("Range") {
            let ranges = range.split(",").collect::<Vec<_>>();
            if ranges.len() == 1 {
                if let Some((start, end)) = parse_range_header(range, self.data.len()) {
                    let sliced_data = self.data.index(start, end);
                    return Ok(Response::build()
                        .header(Header::new(
                            "Content-Range",
                            format!("bytes {}-{}/{}", start, end, self.data.len()),
                        ))
                        .header(Header::new("Accept-Ranges", "bytes"))
                        .header(Header::new("Content-Type", self.content_type.clone()))
                        .header(Header::new("Content-Length", sliced_data.len().to_string()))
                        .status(Status::PartialContent)
                        .streamed_body(Cursor::new(sliced_data.to_vec()))
                        .finalize());
                }
            } else {
                let mut multipart_body: Vec<u8> = Vec::new();
                let boundary = gen_random(32);

                for range in ranges {
                    if let Some((start, end)) = parse_range_header(range, self.data.len()) {
                        let sliced_data = self.data.index(start, end);

                        let mut body: Vec<u8> = Vec::new();

                        body.extend_from_slice(format!("--{boundary}\r\n").as_bytes());
                        body.extend_from_slice(
                            format!(
                                "Content-Range: bytes {}-{}/{}\r\n",
                                start,
                                end,
                                self.data.len()
                            )
                            .as_bytes(),
                        );
                        body.extend_from_slice(
                            format!("Content-Type: {}\r\n\r\n", self.content_type.clone())
                                .as_bytes(),
                        );
                        body.extend_from_slice(&sliced_data);
                        body.extend_from_slice("\r\n".as_bytes());

                        multipart_body.extend_from_slice(&body);
                    }
                }

                multipart_body.extend_from_slice(format!("--{boundary}--\r\n").as_bytes());

                return Ok(Response::build()
                    .header(Header::new("Accept-Ranges", "bytes"))
                    .header(Header::new(
                        "Content-Type",
                        format!("multipart/byteranges; boundary={boundary}"),
                    ))
                    .header(Header::new(
                        "Content-Length",
                        multipart_body.len().to_string(),
                    ))
                    .status(Status::PartialContent)
                    .streamed_body(Cursor::new(multipart_body.to_vec()))
                    .finalize());
            }
        }

        // Add caching headers for static files
        let cache_control_header = self.cache_duration.map_or_else(
            || Header::new("Cache-Control", "no-cache"),
            |duration| Header::new("Cache-Control", format!("public, max-age={duration}")),
        );

        let size = self.data.len();
        let data_full = self.data.full();

        if let Ok(raw) = data_full {
            Ok(Response::build()
                .header(cache_control_header)
                .header(Header::new("Accept-Ranges", "bytes"))
                .header(Header::new("Content-Type", self.content_type))
                .header(Header::new("Content-Length", size.to_string()))
                .streamed_body(Cursor::new(raw))
                .finalize())
        } else {
            Ok(Response::build()
                .header(cache_control_header)
                .header(Header::new("Accept-Ranges", "bytes"))
                .header(Header::new("Content-Type", self.content_type))
                .header(Header::new("Content-Length", size.to_string()))
                .streamed_body(data_full.unwrap_err())
                .finalize())
        }
    }
}

/// Parse the Range header and return the start and end indices.
fn parse_range_header(range: &str, total_len: usize) -> Option<(usize, usize)> {
    if !range.starts_with("bytes=") {
        return None;
    }

    let range = &range[6..];

    if range.starts_with('-') {
        let neg: usize = range.trim_start_matches('-').parse().ok()?;
        return Some((total_len - neg, total_len));
    }

    let parts: Vec<&str> = range.split('-').collect();

    if parts.len() != 2 {
        return Some((parts[0].parse().ok()?, total_len));
    }

    let start = parts[0].parse::<usize>().ok();
    let end = parts[1].parse::<usize>().ok();

    match (start, end) {
        (Some(start), Some(end)) if start <= end && end < total_len => Some((start, end)),
        (Some(start), None) if start < total_len => Some((start, total_len - 1)),
        (None, Some(end)) if end < total_len => Some((0, end)),
        _ => None,
    }
}
