use super::Selector;

pub struct Event {
    modifiers: Vec<String>,
    kind: String,
}

impl Event {
    #[must_use]
    pub fn to_value(&self) -> String {
        if self.kind.starts_with("every") {
            return self.kind.clone();
        }

        format!("{} {}", self.kind, self.modifiers.join(" "))
    }

    /// Add a second event trigger
    #[must_use]
    pub fn and(self, event: Self) -> Trigger {
        Trigger {
            triggers: vec![self, event],
        }
    }

    /// Periodically poll
    ///
    /// Value can be something like `1s [someConditional]`
    #[must_use]
    pub fn poll(value: &str) -> Self {
        Self {
            modifiers: vec![],
            kind: format!("every {value}"),
        }
    }

    /// Standard events refer to web API events (e.g. `click`, `keydown`, `mouseup`, `load`).
    #[allow(clippy::self_named_constructors)]
    #[must_use]
    pub fn event(event: &str) -> Self {
        Self {
            modifiers: vec![],
            kind: event.to_string(),
        }
    }

    /// triggered on load (useful for lazy-loading something)
    #[must_use]
    pub fn on_load() -> Self {
        Self {
            modifiers: vec![],
            kind: "load".to_string(),
        }
    }

    /// triggered when an element is scrolled into the viewport (also useful for lazy-loading). If you are using overflow in css like `overflow-y: scroll` you should use `intersect once` instead of `revealed`.
    #[must_use]
    pub fn on_revealed() -> Self {
        Self {
            modifiers: vec![],
            kind: "revealed".to_string(),
        }
    }

    /// fires once when an element first intersects the viewport. This supports two additional options:
    /// `root:<selector>` - a CSS selector of the root element for intersection
    /// `threshold:<float>` - a floating point number between 0.0 and 1.0, indicating what amount of intersection to fire the event on
    #[must_use]
    pub fn on_intersect(value: &str) -> Self {
        Self {
            modifiers: vec![],
            kind: format!("intersect:{value}"),
        }
    }

    /// the event will only trigger once (e.g. the first click)
    #[must_use]
    pub fn once(mut self) -> Self {
        self.modifiers.push("once".to_string());
        self
    }

    /// the event will only change if the value of the element has changed.
    #[must_use]
    pub fn changed(mut self) -> Self {
        self.modifiers.push("changed".to_string());
        self
    }

    /// a delay will occur before an event triggers a request. If the event is seen again it will reset the delay.
    #[must_use]
    pub fn delay(mut self, delay: &str) -> Self {
        self.modifiers.push(format!("delay:{delay}"));
        self
    }

    /// a throttle will occur after an event triggers a request. If the event is seen again before the delay completes, it is ignored, the element will trigger at the end of the delay.
    #[must_use]
    pub fn throttle(mut self, delay: &str) -> Self {
        self.modifiers.push(format!("throttle:{delay}"));
        self
    }

    /// allows the event that triggers a request to come from another element in the document (e.g. listening to a key event on the body, to support hot keys)
    #[must_use]
    pub fn from(mut self, element: Selector) -> Self {
        self.modifiers.push(format!("from:{}", element.to_value()));
        self
    }

    /// allows you to filter via a CSS selector on the target of the event. This can be useful when you want to listen for triggers from elements that might not be in the DOM at the point of initialization, by, for example, listening on the body, but with a target filter for a child element
    #[must_use]
    pub fn target(mut self, selector: &str) -> Self {
        self.modifiers.push(format!("target:{selector}"));
        self
    }

    /// if this option is included the event will not trigger any other htmx requests on parents (or on elements listening on parents)
    #[must_use]
    pub fn consume(mut self) -> Self {
        self.modifiers.push("consume".to_string());
        self
    }

    /// determines how events are queued if an event occurs while a request for another event is in flight.
    #[must_use]
    pub fn queue(mut self, opt: QueueOption) -> Self {
        self.modifiers.push(format!("queue:{}", opt.to_value()));
        self
    }
}

#[allow(non_camel_case_types)]
#[derive(Default)]
pub enum QueueOption {
    /// queue the first event
    first,
    /// queue the last event (default)
    #[default]
    last,
    /// queue all events (issue a request for each event)
    all,
    /// do not queue new events
    none,
}

impl QueueOption {
    #[must_use]
    pub const fn to_value(&self) -> &str {
        match self {
            Self::first => "first",
            Self::last => "last",
            Self::all => "all",
            Self::none => "none",
        }
    }
}

pub struct Trigger {
    triggers: Vec<Event>,
}

impl From<Event> for Trigger {
    fn from(value: Event) -> Self {
        Self {
            triggers: vec![value],
        }
    }
}

impl Trigger {
    #[must_use]
    pub const fn new() -> Self {
        Self { triggers: vec![] }
    }

    #[must_use]
    pub fn push(mut self, event: Event) -> Self {
        self.triggers.push(event);
        self
    }

    #[must_use]
    pub fn and(self, event: Event) -> Self {
        self.push(event)
    }

    #[must_use]
    pub fn to_value(&self) -> String {
        self.triggers
            .iter()
            .map(Event::to_value)
            .collect::<Vec<_>>()
            .join(", ")
    }
}

impl Default for Trigger {
    fn default() -> Self {
        Self::new()
    }
}
