use maud::{Markup, Render, html};

use crate::{
    auth::{User, csrf::CSRF},
    ui::{AttrExtendable, UIWidget, prelude::HiddenInput},
};

pub enum FormMethod {
    GET,
    POST,
}

pub struct Form {
    action: String,
    method: Option<FormMethod>,
    multipart: bool,
    id: String,
    items: Vec<Box<dyn UIWidget>>,
}

impl Form {
    pub fn new(action: &str) -> Self {
        Self {
            action: action.to_string(),
            method: None,
            multipart: false,
            id: String::new(),
            items: Vec::new(),
        }
    }

    pub fn id(mut self, id: &str) -> Self {
        self.id = id.to_string();
        self
    }

    pub async fn add_csrf(self, u: &User) -> Self {
        self.add_input(
            HiddenInput("csrf", &u.get_csrf().await.to_string()).add_attr("class", "csrf"),
        )
    }

    pub fn multipart(mut self) -> Self {
        self.multipart = true;
        self
    }

    pub fn method(mut self, m: FormMethod) -> Self {
        self.method = Some(m);
        self
    }

    pub fn add_input<T: UIWidget + 'static>(mut self, input: T) -> Self {
        self.items.push(Box::new(input));
        self
    }
}

impl Render for Form {
    fn render(&self) -> maud::Markup {
        self.render_with_class("")
    }
}

impl UIWidget for Form {
    fn can_inherit(&self) -> bool {
        true
    }

    fn base_class(&self) -> Vec<String> {
        Vec::new()
    }

    fn extended_class(&self) -> Vec<String> {
        Vec::new()
    }

    fn render_with_class(&self, class: &str) -> Markup {
        let method = self
            .method
            .as_ref()
            .map(|x| match x {
                FormMethod::GET => "get",
                FormMethod::POST => "post",
            })
            .unwrap_or("post");

        html! {
                form action=(self.action) method=(method) id=(self.id) enctype=(if self.multipart { "multipart/form-data" } else { "application/x-www-form-urlencoded" }) class=(class) {
                @for item in &self.items  {
                    (item)
                };
            }
        }
    }
}
