use crate::ui::{
    UIWidget,
    components::prelude::{Classic, ClassicWidget},
};
use maud::{Markup, PreEscaped, Render, html};

#[allow(non_snake_case)]
#[must_use]
pub fn OrderedList() -> ListWidget {
    ListWidget(Vec::new(), true)
}

#[allow(non_snake_case)]
#[must_use]
pub fn UnorderedList() -> ListWidget {
    ListWidget(Vec::new(), false)
}

#[allow(non_snake_case)]
#[must_use]
pub fn HorizontalList(list: ListWidget) -> ClassicWidget<ListWidget> {
    Classic(
        "flex flex-wrap items-center justify-center text-gray-900 dark:text-white",
        list,
    )
}

pub struct ListWidget(Vec<Box<dyn UIWidget>>, bool);

impl ListWidget {
    #[must_use]
    pub fn push<T: UIWidget + 'static>(mut self, element: T) -> Self {
        self.0.push(Box::new(element));
        self
    }

    #[must_use]
    pub fn push_some<T: UIWidget + 'static, X, U: Fn(X) -> T>(
        mut self,
        option: Option<X>,
        then: U,
    ) -> Self {
        if let Some(val) = option {
            self.0.push(Box::new(then(val)));
        }
        self
    }

    #[must_use]
    pub fn push_if<T: UIWidget + 'static, U: Fn() -> T>(
        mut self,
        condition: bool,
        then: U,
    ) -> Self {
        if condition {
            self.0.push(Box::new(then()));
        }
        self
    }

    #[must_use]
    pub fn push_for_each<T, X, F>(mut self, items: &[X], mut action: F) -> Self
    where
        T: UIWidget + 'static,
        F: FnMut(&X) -> T,
    {
        for item in items {
            self.0.push(Box::new(action(item)));
        }

        self
    }
}

impl Render for ListWidget {
    fn render(&self) -> Markup {
        self.render_with_class("")
    }
}

impl UIWidget for ListWidget {
    fn can_inherit(&self) -> bool {
        true
    }

    fn base_class(&self) -> Vec<String> {
        vec![]
    }

    fn extended_class(&self) -> Vec<String> {
        vec![]
    }

    fn render_with_class(&self, class: &str) -> Markup {
        let inner = html! {
            @for e in &self.0 {
                li { (e.as_ref()) };
            }
        };

        if self.1 {
            html! {
                ol class=(class) {
                    (inner);
                }
            }
        } else {
            html! {
                ul class=(class) {
                    (inner);
                }
            }
        }
    }
}

// TODO : List data backed list + reorderable + add + remove + crud

#[allow(non_snake_case)]
pub fn CheckIconRounded() -> PreEscaped<String> {
    html! {
    svg class="w-3.5 h-3.5 me-2 text-green-500 dark:text-green-400 shrink-0" aria-hidden="true" xmlns="http://www.w3.org/2000/svg" fill="currentColor" viewBox="0 0 20 20" {
    path d="M10 .5a9.5 9.5 0 1 0 9.5 9.5A9.51 9.51 0 0 0 10 .5Zm3.707 8.207-4 4a1 1 0 0 1-1.414 0l-2-2a1 1 0 0 1 1.414-1.414L9 10.586l3.293-3.293a1 1 0 0 1 1.414 1.414Z" {};
    };
    }
}

#[allow(non_snake_case)]
pub fn CheckIconRoundedGray() -> PreEscaped<String> {
    html! {
        svg class="w-3.5 h-3.5 me-2 text-gray-500 dark:text-gray-400 shrink-0" aria-hidden="true" xmlns="http://www.w3.org/2000/svg" fill="currentColor" viewBox="0 0 20 20" {
            path d="M10 .5a9.5 9.5 0 1 0 9.5 9.5A9.51 9.51 0 0 0 10 .5Zm3.707 8.207-4 4a1 1 0 0 1-1.414 0l-2-2a1 1 0 0 1 1.414-1.414L9 10.586l3.293-3.293a1 1 0 0 1 1.414 1.414Z" {}
    };
    }
}

#[allow(non_snake_case)]
pub fn ListEntry<T: UIWidget + 'static, I: UIWidget + 'static>(
    icon: I,
    inner: T,
) -> PreEscaped<String> {
    html! {
        li class="flex items-center" {
        (icon)
        (inner)
    };
    }
}

#[allow(non_snake_case)]
pub fn CheckIcon() -> PreEscaped<String> {
    html! {
        svg class="shrink-0 w-3.5 h-3.5 text-green-500 dark:text-green-400" aria-hidden="true" xmlns="http://www.w3.org/2000/svg" fill="none" viewBox="0 0 16 12" {
            path stroke="currentColor" stroke-linecap="round" stroke-linejoin="round" stroke-width="2" d="M1 5.917 5.724 10.5 15 1.5" {};
    };
    }
}
