use crate::ui::{UIWidget, color::UIColor};
use maud::{Markup, PreEscaped, Render, html};

use super::{Nothing, space::ScreenValue};

#[allow(non_snake_case)]
/// Text UI Widget
#[must_use]
pub fn Text(txt: &str) -> TextWidget {
    TextWidget {
        inner: None,
        txt: txt.to_string(),
        family: String::new(),
        title: None,
        font: String::new(),
        color: String::new(),
        style: Vec::new(),
        size: String::new(),
        line_height: None,
        overflow: None,
        wrap: None,
        indent: None,
        transform: None,
        decoration: None,
        whitespace: None,
        wordbreak: None,
        hyphens: None,
        spacing: None,
        clamp: None,
        pseudo: None,
        align: None,
        vert_align: None,
        list_style: None,
        select: None,
        kind: TextKind::Paragraph,
    }
}

enum TextKind {
    Paragraph,
    Span,
    Pre,
}

#[allow(non_snake_case)]
/// HTML `<p>` Paragraph
pub fn Paragraph<T: UIWidget + 'static>(inner: T) -> TextWidget {
    TextWidget {
        inner: Some(Box::new(inner)),
        title: None,
        font: String::new(),
        family: String::new(),
        color: String::new(),
        txt: String::new(),
        style: Vec::new(),
        size: String::new(),
        indent: None,
        overflow: None,
        decoration: None,
        whitespace: None,
        wordbreak: None,
        hyphens: None,
        wrap: None,
        spacing: None,
        transform: None,
        pseudo: None,
        vert_align: None,
        line_height: None,
        list_style: None,
        clamp: None,
        align: None,
        select: None,
        kind: TextKind::Paragraph,
    }
}

#[allow(non_snake_case)]
/// `<span>` element
#[must_use]
pub fn Span(txt: &str) -> TextWidget {
    TextWidget {
        inner: None,
        title: None,
        txt: txt.to_string(),
        family: String::new(),
        font: String::new(),
        style: Vec::new(),
        color: String::new(),
        list_style: None,
        size: String::new(),
        indent: None,
        overflow: None,
        whitespace: None,
        wordbreak: None,
        hyphens: None,
        decoration: None,
        transform: None,
        wrap: None,
        vert_align: None,
        spacing: None,
        line_height: None,
        clamp: None,
        align: None,
        pseudo: None,
        select: None,
        kind: TextKind::Span,
    }
}

#[allow(non_snake_case)]
/// `<pre>` element
#[must_use]
pub fn Code(txt: &str) -> TextWidget {
    TextWidget {
        inner: None,
        title: None,
        txt: txt.to_string(),
        family: String::new(),
        font: String::new(),
        style: Vec::new(),
        color: String::new(),
        list_style: None,
        size: String::new(),
        indent: None,
        overflow: None,
        whitespace: None,
        wordbreak: None,
        hyphens: None,
        decoration: None,
        transform: None,
        wrap: None,
        vert_align: None,
        spacing: None,
        line_height: None,
        clamp: None,
        align: None,
        pseudo: None,
        select: None,
        kind: TextKind::Pre,
    }
}

pub struct TextWidget {
    inner: Option<Box<dyn UIWidget>>,
    txt: String,
    family: String,
    style: Vec<String>,
    font: String,
    color: String,
    list_style: Option<ListStyle>,
    line_height: Option<LineHeight>,
    decoration: Option<DecorationWidget>,
    transform: Option<TextTransform>,
    vert_align: Option<VerticalTextAlignment>,
    overflow: Option<TextOverflow>,
    indent: Option<ScreenValue>,
    wrap: Option<TextWrap>,
    whitespace: Option<TextWhitespace>,
    wordbreak: Option<TextWordBreak>,
    hyphens: Option<TextHyphens>,
    size: String,
    kind: TextKind,
    spacing: Option<LetterSpacing>,
    pseudo: Option<TextContent>,
    align: Option<TextAlignment>,
    clamp: Option<LineClamp>,
    select: Option<TextSelection>,
    title: Option<String>,
}

impl TextWidget {
    #[must_use]
    pub fn select(mut self, select: TextSelection) -> Self {
        self.select = Some(select);
        self
    }

    #[must_use]
    pub fn whitespace(mut self, whitespace: TextWhitespace) -> Self {
        self.whitespace = Some(whitespace);
        self
    }

    #[must_use]
    pub fn wordbreak(mut self, wordbreak: TextWordBreak) -> Self {
        self.wordbreak = Some(wordbreak);
        self
    }

    #[must_use]
    pub fn hyphen(mut self, hyphen: TextHyphens) -> Self {
        self.hyphens = Some(hyphen);
        self
    }

    #[must_use]
    pub fn underlined(self) -> Self {
        self.decoration(TextDecoration(DecorationKind::Underline))
    }

    #[must_use]
    pub fn overlined(self) -> Self {
        self.decoration(TextDecoration(DecorationKind::Overline))
    }

    #[must_use]
    pub fn strikethrough(self) -> Self {
        self.decoration(TextDecoration(DecorationKind::LineThrough))
    }

    // Weight

    #[must_use]
    pub fn thin(mut self) -> Self {
        self.font = "font-thin".to_string();
        self
    }

    #[must_use]
    pub fn extralight(mut self) -> Self {
        self.font = "font-extralight".to_string();
        self
    }

    #[must_use]
    pub fn light(mut self) -> Self {
        self.font = "font-light".to_string();
        self
    }

    #[must_use]
    pub fn normal(mut self) -> Self {
        self.font = "font-normal".to_string();
        self
    }

    /// Turn `Text` medium.
    ///
    /// Adds the class `font-medium`
    #[must_use]
    pub fn medium(mut self) -> Self {
        self.font = "font-medium".to_string();
        self
    }

    /// Turn `Text` semibold.
    ///
    /// Adds the class `font-semibold`
    #[must_use]
    pub fn semibold(mut self) -> Self {
        self.font = "font-semibold".to_string();
        self
    }

    /// Turn `Text` bold.
    ///
    /// Adds the class `font-bold`
    #[must_use]
    pub fn bold(mut self) -> Self {
        self.font = "font-bold".to_string();
        self
    }

    #[must_use]
    pub fn extrabold(mut self) -> Self {
        self.font = "font-extrabold".to_string();
        self
    }

    #[must_use]
    pub fn weight_black(mut self) -> Self {
        self.font = "font-black".to_string();
        self
    }

    // Styles

    #[must_use]
    pub fn italic(mut self, apply: bool) -> Self {
        if apply {
            self.style.push("italic".to_string());
        } else {
            self.style.push("not-italic".to_string())
        }
        self
    }

    #[must_use]
    pub fn wrap(mut self, wrap: TextWrap) -> Self {
        self.wrap = Some(wrap);
        self
    }

    #[must_use]
    pub fn indentation(mut self, indent: ScreenValue) -> Self {
        self.indent = Some(indent);
        self
    }

    #[must_use]
    pub fn align(mut self, alignment: TextAlignment) -> Self {
        self.align = Some(alignment);
        self
    }

    #[must_use]
    pub fn align_vertical(mut self, alignment: VerticalTextAlignment) -> Self {
        self.vert_align = Some(alignment);
        self
    }

    #[must_use]
    pub fn transform(mut self, transform: TextTransform) -> Self {
        self.transform = Some(transform);
        self
    }

    #[must_use]
    pub fn number_style(mut self, s: NumberStyle) -> Self {
        self.style.push(s.to_value().to_string());
        self
    }

    #[must_use]
    pub fn overflow(mut self, overflow: TextOverflow) -> Self {
        self.overflow = Some(overflow);
        self
    }

    #[must_use]
    pub fn max_lines(mut self, l: LineClamp) -> Self {
        self.clamp = Some(l);
        self
    }

    #[must_use]
    pub fn decoration(mut self, decoration: DecorationWidget) -> Self {
        self.decoration = Some(decoration);
        self
    }

    #[must_use]
    pub fn list_style(mut self, style: ListStyle) -> Self {
        self.list_style = Some(style);
        self
    }

    #[must_use]
    pub fn content(mut self, content: TextContent) -> Self {
        self.pseudo = Some(content);
        self
    }

    #[must_use]
    pub fn line_height(mut self, height: LineHeight) -> Self {
        self.line_height = Some(height);
        self
    }

    #[must_use]
    pub fn title(mut self, title: &str) -> Self {
        self.title = Some(title.replace('\'', "\\'"));
        self
    }

    // Sizes

    #[must_use]
    pub fn _9xl(mut self) -> Self {
        self.size = "text-9xl".to_string();
        self
    }

    #[must_use]
    pub fn _8xl(mut self) -> Self {
        self.size = "text-8xl".to_string();
        self
    }

    #[must_use]
    pub fn _7xl(mut self) -> Self {
        self.size = "text-7xl".to_string();
        self
    }

    #[must_use]
    pub fn _6xl(mut self) -> Self {
        self.size = "text-6xl".to_string();
        self
    }

    #[must_use]
    pub fn _5xl(mut self) -> Self {
        self.size = "text-5xl".to_string();
        self
    }

    #[must_use]
    pub fn _4xl(mut self) -> Self {
        self.size = "text-4xl".to_string();
        self
    }

    #[must_use]
    pub fn large(mut self) -> Self {
        self.size = "text-lg".to_string();
        self
    }

    #[must_use]
    pub fn base_size(mut self) -> Self {
        self.size = "text-base".to_string();
        self
    }

    /// Turn `Text` size to 3XL.
    ///
    /// Adds the class `text-3xl`
    #[must_use]
    pub fn _3xl(mut self) -> Self {
        self.size = "text-3xl".to_string();
        self
    }

    /// Turn `Text` size to 2XL.
    ///
    /// Adds the class `text-2xl`
    #[must_use]
    pub fn _2xl(mut self) -> Self {
        self.size = "text-2xl".to_string();
        self
    }

    /// Turn `Text` size to xl.
    ///
    /// Adds the class `text-xl`
    #[must_use]
    pub fn xl(mut self) -> Self {
        self.size = "text-xl".to_string();
        self
    }

    /// Turn `Text` size to small.
    ///
    /// Adds the class `text-sm`
    #[must_use]
    pub fn sm(mut self) -> Self {
        self.size = "text-sm".to_string();
        self
    }

    /// Turn `Text` size to x small.
    ///
    /// Adds the class `text-xs`
    #[must_use]
    pub fn xs(mut self) -> Self {
        self.size = "text-xs".to_string();
        self
    }

    // Text Color

    #[must_use]
    pub fn color<T: UIColor>(mut self, color: &T) -> Self {
        self.color = format!("text-{}", color.color_class());
        self
    }

    #[must_use]
    pub fn black(mut self) -> Self {
        self.color = "text-black".to_string();
        self
    }

    #[must_use]
    pub fn white(mut self) -> Self {
        self.color = "text-white".to_string();
        self
    }

    // Font Family

    #[must_use]
    pub fn sans(mut self) -> Self {
        self.family = "font-sans".to_string();
        self
    }

    #[must_use]
    pub fn serif(mut self) -> Self {
        self.family = "font-serif".to_string();
        self
    }

    #[must_use]
    pub fn mono(mut self) -> Self {
        self.family = "font-mono".to_string();
        self
    }
}

impl Render for TextWidget {
    fn render(&self) -> Markup {
        self.render_with_class("")
    }
}

impl UIWidget for TextWidget {
    fn can_inherit(&self) -> bool {
        true
    }

    fn base_class(&self) -> Vec<String> {
        let mut ret = vec![
            self.color.clone(),
            self.font.clone(),
            self.size.clone(),
            self.family.clone(),
        ];

        macro_rules! add_option {
            ($opt:ident, $ret:ident) => {
                if let Some($opt) = &self.$opt {
                    $ret.push($opt.to_value().to_string());
                }
            };
        }

        add_option!(spacing, ret);

        if let Some(indent) = &self.indent {
            ret.push(format!("indent-{}", indent.to_value()));
        }

        add_option!(clamp, ret);
        add_option!(align, ret);
        add_option!(vert_align, ret);
        add_option!(list_style, ret);
        add_option!(pseudo, ret);
        add_option!(line_height, ret);
        add_option!(select, ret);

        if let Some(decoration) = &self.decoration {
            ret.extend_from_slice(&decoration.base_class());
        }

        ret.extend_from_slice(&self.style);

        add_option!(transform, ret);
        add_option!(overflow, ret);
        add_option!(wrap, ret);
        add_option!(whitespace, ret);
        add_option!(wordbreak, ret);
        add_option!(hyphens, ret);

        ret
    }

    fn extended_class(&self) -> Vec<String> {
        let mut c = self.base_class();
        if let Some(inner) = &self.inner {
            c.extend_from_slice(&inner.extended_class());
        }
        c
    }

    fn render_with_class(&self, class: &str) -> Markup {
        let element = match self.kind {
            TextKind::Paragraph => "p",
            TextKind::Span => "span",
            TextKind::Pre => "pre",
        };

        let mut ret = "<".to_string();
        ret.push_str(&element);
        ret.push_str(" class='");
        ret.push_str(&format!("{} {}", class, self.base_class().join(" ")));
        ret.push_str("'");

        if let Some(title) = &self.title {
            ret.push_str(" title='");
            ret.push_str(&title);
            ret.push_str("'");
        }

        ret.push_str(">");

        if let Some(inner) = &self.inner {
            ret.push_str(&inner.render().0);
        } else {
            ret.push_str(&self.txt);
        }

        ret.push_str("</");
        ret.push_str(&element);
        ret.push_str(">");

        PreEscaped(ret)
    }
}

pub enum NumberStyle {
    Normal,
    Ordinal,
    SlashedZero,
    OldStyle,
    Lining,
    Proportional,
    Tabular,
    DiagonalFractions,
    StackedFractions,
}

impl NumberStyle {
    pub const fn to_value(&self) -> &str {
        match self {
            NumberStyle::Normal => "normal-nums",
            NumberStyle::Ordinal => "ordinal",
            NumberStyle::SlashedZero => "slashed-zero",
            NumberStyle::OldStyle => "oldstyle-nums",
            NumberStyle::Lining => "lining-nums",
            NumberStyle::Proportional => "proportional-nums",
            NumberStyle::Tabular => "tabular-nums",
            NumberStyle::DiagonalFractions => "diagonal-fractions",
            NumberStyle::StackedFractions => "stacked-fractions",
        }
    }
}

pub enum LetterSpacing {
    Tighter,
    Tight,
    Normal,
    Wide,
    Wider,
    Widest,
}

impl LetterSpacing {
    pub const fn to_value(&self) -> &str {
        match self {
            LetterSpacing::Tighter => "tracking-tighter",
            LetterSpacing::Tight => "tracking-tight",
            LetterSpacing::Normal => "tracking-normal",
            LetterSpacing::Wide => "tracking-wide",
            LetterSpacing::Wider => "tracking-wider",
            LetterSpacing::Widest => "tracking-widest",
        }
    }
}

pub enum LineClamp {
    None,
    _1,
    _2,
    _3,
    _4,
    _5,
    _6,
}

impl LineClamp {
    pub fn to_value(&self) -> &str {
        match self {
            LineClamp::None => "line-clamp-none",
            LineClamp::_1 => "line-clamp-1",
            LineClamp::_2 => "line-clamp-2",
            LineClamp::_3 => "line-clamp-3",
            LineClamp::_4 => "line-clamp-4",
            LineClamp::_5 => "line-clamp-5",
            LineClamp::_6 => "line-clamp-6",
        }
    }
}

pub enum TextAlignment {
    Left,
    Center,
    Right,
    Justify,
    Start,
    End,
}

impl TextAlignment {
    pub const fn to_value(&self) -> &str {
        match self {
            TextAlignment::Left => "text-left",
            TextAlignment::Center => "text-center",
            TextAlignment::Right => "text-right",
            TextAlignment::Justify => "text-justify",
            TextAlignment::Start => "text-start",
            TextAlignment::End => "text-end",
        }
    }
}

pub enum ListStyle {
    Url(String),
    None,
    Disc,
    Decimal,
}

impl ListStyle {
    pub fn to_value(&self) -> String {
        match self {
            ListStyle::Url(url) => format!("list-image-[url({url})]"),
            ListStyle::None => "list-none".to_string(),
            ListStyle::Disc => "list-disc".to_string(),
            ListStyle::Decimal => "list-decimal".to_string(),
        }
    }
}

pub enum TextContent {
    None,
    Before(String),
    After(String),
}

impl TextContent {
    pub fn to_value(&self) -> String {
        match self {
            TextContent::None => "content-none".to_string(),
            TextContent::Before(c) => format!("before:content-['{c}']"),
            TextContent::After(c) => format!("after:content-['{c}']"),
        }
    }
}

pub enum LineHeight {
    _3,
    _4,
    _5,
    _6,
    _7,
    _8,
    _9,
    _10,
    None,
    Tight,
    Snug,
    Normal,
    Relaxed,
    Loose,
}

impl LineHeight {
    pub const fn to_value(&self) -> &str {
        match self {
            LineHeight::_3 => "leading-3",
            LineHeight::_4 => "leading-4",
            LineHeight::_5 => "leading-5",
            LineHeight::_6 => "leading-6",
            LineHeight::_7 => "leading-7",
            LineHeight::_8 => "leading-8",
            LineHeight::_9 => "leading-9",
            LineHeight::_10 => "leading-10",
            LineHeight::None => "leading-none",
            LineHeight::Tight => "leading-tight",
            LineHeight::Snug => "leading-snug",
            LineHeight::Normal => "leading-normal",
            LineHeight::Relaxed => "leading-relaxed",
            LineHeight::Loose => "leading-loose",
        }
    }
}

// Decoration

#[allow(non_snake_case)]
pub fn TextDecoration(kind: DecorationKind) -> DecorationWidget {
    DecorationWidget {
        kind: kind.to_value().to_string(),
        color: None,
        style: None,
        underline_offset: None,
        thickness: None,
    }
}

pub struct DecorationWidget {
    kind: String,
    color: Option<String>,
    style: Option<DecorationStyle>,
    thickness: Option<DecorationThickness>,
    underline_offset: Option<UnderlineOffset>,
}

impl DecorationWidget {
    #[must_use]
    pub fn thickness(mut self, thickness: DecorationThickness) -> Self {
        self.thickness = Some(thickness);
        self
    }

    #[must_use]
    pub fn style(mut self, style: DecorationStyle) -> Self {
        self.style = Some(style);
        self
    }

    #[must_use]
    pub fn color<C: UIColor>(mut self, color: C) -> Self {
        self.color = Some(format!("decoration-{}", color.color_class()));
        self
    }

    #[must_use]
    pub fn underline_offset(mut self, offset: UnderlineOffset) -> Self {
        self.underline_offset = Some(offset);
        self
    }
}

pub enum DecorationKind {
    Underline,
    Overline,
    LineThrough,
    NoUnderline,
}

impl DecorationKind {
    pub const fn to_value(&self) -> &str {
        match self {
            DecorationKind::Underline => "underline",
            DecorationKind::Overline => "overline",
            DecorationKind::LineThrough => "line-through",
            DecorationKind::NoUnderline => "no-underline",
        }
    }
}

pub enum DecorationThickness {
    Auto,
    FromFont,
    _0,
    _1,
    _2,
    _4,
    _8,
}

impl DecorationThickness {
    pub const fn to_value(&self) -> &str {
        match self {
            DecorationThickness::Auto => "decoration-auto",
            DecorationThickness::FromFont => "decoration-from-font",
            DecorationThickness::_0 => "decoration-0",
            DecorationThickness::_1 => "decoration-1",
            DecorationThickness::_2 => "decoration-2",
            DecorationThickness::_4 => "decoration-4",
            DecorationThickness::_8 => "decoration-8",
        }
    }
}

pub enum DecorationStyle {
    Solid,
    Double,
    Dotted,
    Dashed,
    Wavy,
}

impl DecorationStyle {
    pub const fn to_value(&self) -> &str {
        match self {
            DecorationStyle::Solid => "decoration-solid",
            DecorationStyle::Double => "decoration-double",
            DecorationStyle::Dotted => "decoration-dotted",
            DecorationStyle::Dashed => "decoration-dashed",
            DecorationStyle::Wavy => "decoration-wavy",
        }
    }
}

impl Render for DecorationWidget {
    fn render(&self) -> Markup {
        self.render_with_class("")
    }
}

impl UIWidget for DecorationWidget {
    fn can_inherit(&self) -> bool {
        false
    }

    fn base_class(&self) -> Vec<String> {
        let mut ret = vec![self.kind.clone()];

        if let Some(color) = &self.color {
            ret.push(color.clone());
        }

        if let Some(style) = &self.style {
            ret.push(style.to_value().to_string());
        }

        if let Some(thickness) = &self.thickness {
            ret.push(thickness.to_value().to_string());
        }

        if let Some(offset) = &self.underline_offset {
            ret.push(offset.to_value().to_string());
        }

        ret
    }

    fn extended_class(&self) -> Vec<String> {
        self.base_class()
    }

    fn render_with_class(&self, _: &str) -> Markup {
        Nothing()
    }
}

pub enum UnderlineOffset {
    Auto,
    _0,
    _1,
    _2,
    _4,
    _8,
}

impl UnderlineOffset {
    pub const fn to_value(&self) -> &str {
        match self {
            Self::Auto => "underline-offset-auto",
            Self::_0 => "underline-offset-0",
            Self::_1 => "underline-offset-1",
            Self::_2 => "underline-offset-2",
            Self::_4 => "underline-offset-4",
            Self::_8 => "underline-offset-8",
        }
    }
}

pub enum VerticalTextAlignment {
    Baseline,
    Top,
    Middle,
    Bottom,
    TextTop,
    TextBottom,
    Sub,
    Super,
}

impl VerticalTextAlignment {
    pub const fn to_value(&self) -> &str {
        match self {
            VerticalTextAlignment::Baseline => "align-baseline",
            VerticalTextAlignment::Top => "align-top",
            VerticalTextAlignment::Middle => "align-middle",
            VerticalTextAlignment::Bottom => "align-bottom",
            VerticalTextAlignment::TextTop => "align-text-top",
            VerticalTextAlignment::TextBottom => "align-text-bottom",
            VerticalTextAlignment::Sub => "align-sub",
            VerticalTextAlignment::Super => "align-super",
        }
    }
}

pub enum TextTransform {
    Uppecase,
    Lowercase,
    Capitalize,
    None,
}

impl TextTransform {
    pub const fn to_value(&self) -> &str {
        match self {
            TextTransform::Uppecase => "uppercase",
            TextTransform::Lowercase => "lowercase",
            TextTransform::Capitalize => "capitalize",
            TextTransform::None => "normal-case",
        }
    }
}

pub enum TextOverflow {
    Truncate,
    Ellipsis,
    Clip,
}

impl TextOverflow {
    pub const fn to_value(&self) -> &str {
        match self {
            TextOverflow::Truncate => "truncate",
            TextOverflow::Ellipsis => "text-ellipsis",
            TextOverflow::Clip => "text-clip",
        }
    }
}

pub enum TextWrap {
    Wrap,
    NoWrap,
    Balance,
    Pretty,
}

impl TextWrap {
    pub const fn to_value(&self) -> &str {
        match self {
            TextWrap::Wrap => "text-wrap",
            TextWrap::NoWrap => "text-nowrap",
            TextWrap::Balance => "text-balance",
            TextWrap::Pretty => "text-pretty",
        }
    }
}

pub enum TextWhitespace {
    Normal,
    NoWrap,
    Pre,
    PreLine,
    PreWrap,
    BreakSpaces,
}

impl TextWhitespace {
    pub const fn to_value(&self) -> &str {
        match self {
            TextWhitespace::Normal => "whitespace-normal",
            TextWhitespace::NoWrap => "whitespace-nowrap",
            TextWhitespace::Pre => "whitespace-pre",
            TextWhitespace::PreLine => "whitespace-pre-line",
            TextWhitespace::PreWrap => "whitespace-pre-wrap",
            TextWhitespace::BreakSpaces => "whitespace-break-spaces",
        }
    }
}

pub enum TextWordBreak {
    Normal,
    Words,
    All,
    Keep,
}

impl TextWordBreak {
    pub const fn to_value(&self) -> &str {
        match self {
            TextWordBreak::Normal => "break-normal",
            TextWordBreak::Words => "break-words",
            TextWordBreak::All => "break-all",
            TextWordBreak::Keep => "break-keep",
        }
    }
}

pub enum TextHyphens {
    None,
    Manual,
    Auto,
}

impl TextHyphens {
    pub const fn to_value(&self) -> &str {
        match self {
            TextHyphens::None => "hyphens-none",
            TextHyphens::Manual => "hyphens-manual",
            TextHyphens::Auto => "hyphens-auto",
        }
    }
}

pub enum TextSelection {
    None,
    Text,
    All,
    Auto,
}

impl TextSelection {
    pub const fn to_value(&self) -> &str {
        match self {
            TextSelection::None => "select-none",
            TextSelection::Text => "select-text",
            TextSelection::All => "select-all",
            TextSelection::Auto => "select-auto	",
        }
    }
}

macro_rules! color_widget {
    ($constr:ident, $widget:ident, $class:literal) => {
        #[allow(non_snake_case)]
        pub fn $constr<T: UIWidget + 'static, C: UIColor + 'static>(color: C, inner: T) -> $widget {
            $widget(Box::new(inner), Box::new(color))
        }

        pub struct $widget(Box<dyn UIWidget>, Box<dyn UIColor>);

        impl Render for $widget {
            fn render(&self) -> Markup {
                self.render_with_class("")
            }
        }

        impl UIWidget for $widget {
            fn can_inherit(&self) -> bool {
                false
            }

            fn base_class(&self) -> Vec<String> {
                let mut class = $class.to_string();
                class.push_str(&format!("-{}", self.1.color_class()));
                vec![class]
            }

            fn extended_class(&self) -> Vec<String> {
                let mut c = self.base_class();
                c.extend_from_slice(&self.0.extended_class());
                c
            }

            fn render_with_class(&self, class: &str) -> Markup {
                if self.0.as_ref().can_inherit() {
                    self.0
                        .as_ref()
                        .render_with_class(&format!("{} {class}", self.base_class().join(" ")))
                } else {
                    html! {
                        div class=(format!("{} {class}", self.base_class().join(" "))) {
                            (self.0.as_ref())
                        }
                    }
                }
            }
        }
    };
}

color_widget!(TextCursorColor, CaretColorWidget, "caret");
color_widget!(AccentColor, AccentColorWidget, "accent");

// TODO : markdown render widget
